import importlib
import code
import logging
import os
import string
import sys
import time
import unicodedata
from typing import Any, Optional


# Sentinel value for undefined argument
UNDEFINED = object()


def delay(sleeptime: float = 0.0):
    """Delay execution for given amount of seconds.

    :param sleeptime: seconds as float, defaults to 0
    """
    if delay is None:
        return

    sleeptime = float(sleeptime)

    if sleeptime > 0:
        logging.debug("Sleeping for %f second(s)", sleeptime)
        time.sleep(sleeptime)


def clean_filename(filename: str, replace: str = " ") -> str:
    """Clean filename to valid format which can be used file operations.

    :param filename: name to be cleaned
    :param replace: characters to replace with underscore, defaults to " "
    :return: valid filename
    """
    valid_characters = "-_.()" + string.ascii_letters + string.digits

    for char in replace:
        filename = filename.replace(char, "_")

    clean = unicodedata.normalize("NFKD", filename)
    clean = clean.encode("ASCII", "ignore").decode()
    clean = "".join(char for char in filename if char in valid_characters)

    return clean


def required_env(name: str, default: Any = UNDEFINED) -> str:
    """Load required environment variable.

    :param name: Name of environment variable
    :param default: Value to use if variable is undefined.
                    If not given and variable is undefined, raises KeyError.
    """
    val = os.getenv(name, default)
    if val is UNDEFINED:
        raise KeyError(f"Missing required environment variable: {name}")
    return val


def required_param(param_name: Any = None, method_name: str = None):
    """Check that required parameter is not None"""
    if not isinstance(param_name, list):
        param_name = [param_name]
    if any(p is None for p in param_name):
        raise KeyError("Required parameter(s) missing for kw: %s" % method_name)


def import_by_name(name: str, caller: str = None) -> Any:
    """Import module (or attribute) by name.

    :param name: Import path, e.g. RPA.Robocorp.WorkItems.RobocorpAdapter
    """
    name = str(name)

    # Attempt import as path module
    try:
        return importlib.import_module(name)
    except ImportError:
        pass

    # Attempt import from calling file
    if caller is not None:
        try:
            module = importlib.import_module(caller)
            return getattr(module, name)
        except AttributeError:
            pass

    # Attempt import as path to attribute inside module
    if "." in name:
        try:
            path, attr = name.rsplit(".", 1)
            module = importlib.import_module(path)
            return getattr(module, attr)
        except (AttributeError, ImportError):
            pass

    raise ValueError(f"No module/attribute with name: {name}")


def interact(expression: Any = None, local: Optional[dict] = None):
    """Interrupts the execution with an interactive shell on `expression`."""
    if expression is not None and not expression:
        return

    sys.stdin, bkp_stdin = sys.__stdin__, sys.stdin
    sys.stdout, bkp_stdout = sys.__stdout__, sys.stdout
    sys.stderr, bkp_stderr = sys.__stderr__, sys.stderr
    try:
        code.interact(local=local or {**globals(), **locals()})
    finally:
        sys.stdin = bkp_stdin
        sys.stdout = bkp_stdout
        sys.stderr = bkp_stderr



from dataclasses import dataclass, astuple
from typing import Any, Optional, Union, Sequence, Tuple


def to_point(obj: Any) -> Optional["Point"]:
    """Convert `obj` to instance of Point."""
    if obj is None or isinstance(obj, Point):
        return obj
    if isinstance(obj, str):
        obj = obj.split(",")
    return Point(*(int(i) for i in obj))


def to_region(obj: Any) -> Optional["Region"]:
    """Convert `obj` to instance of Region."""
    if obj is None or isinstance(obj, Region):
        return obj
    if isinstance(obj, str):
        obj = obj.split(",")
    return Region(*(int(i) for i in obj))


@dataclass
class Undefined:
    """Internal placeholder for generic geometry."""

    def __str__(self):
        return "undefined"


@dataclass(order=True)
class Point:
    """Container for a 2D point."""

    x: int
    y: int

    def __post_init__(self):
        self.x = int(self.x)
        self.y = int(self.y)

    def __str__(self):
        return f"point:{self.x},{self.y}"

    def __iter__(self):
        return iter(self.as_tuple())

    def as_tuple(self) -> Tuple:
        return astuple(self)

    def move(self, x: int, y: int) -> "Point":
        """Move the point relativce to the current position,
        and return the resulting copy.
        """
        return Point(self.x + int(x), self.y + int(y))


@dataclass(order=True)
class Region:
    """Container for a 2D rectangular region."""

    left: int
    top: int
    right: int
    bottom: int

    def __post_init__(self):
        self.left = int(self.left)
        self.top = int(self.top)
        self.right = int(self.right)
        self.bottom = int(self.bottom)

        if self.left >= self.right:
            raise ValueError("Invalid width")
        if self.top >= self.bottom:
            raise ValueError("Invalid height")

    def __str__(self):
        return f"region:{self.left},{self.top},{self.right},{self.bottom}"

    def __iter__(self):
        return iter(self.as_tuple())

    @classmethod
    def from_size(cls, left: int, top: int, width: int, height: int) -> "Region":
        return cls(left, top, left + width, top + height)

    @classmethod
    def merge(cls, regions: Sequence["Region"]) -> "Region":
        left = min(region.left for region in regions)
        top = min(region.top for region in regions)
        right = max(region.right for region in regions)
        bottom = max(region.bottom for region in regions)

        return cls(left, top, right, bottom)

    @property
    def width(self) -> int:
        return self.right - self.left

    @width.setter
    def width(self, value: int):
        diff = int(value) - self.width
        if self.width + diff <= 0:
            raise ValueError("Invalid width")

        self.left -= int(diff / 2)
        self.right += int(diff / 2)

    @property
    def height(self) -> int:
        return self.bottom - self.top

    @height.setter
    def height(self, value: int):
        diff = int(value) - self.height
        if self.height + diff <= 0:
            raise ValueError("Invalid height")

        self.top -= int(diff / 2)
        self.bottom += int(diff / 2)

    @property
    def area(self) -> int:
        return self.width * self.height

    @property
    def center(self) -> Point:
        return Point(
            x=int((self.left + self.right) / 2), y=int((self.top + self.bottom) / 2)
        )

    def as_tuple(self) -> Tuple:
        return astuple(self)

    def scale(self, scaling_factor: float) -> "Region":
        """Scale all coordinate values with a given factor.

        Used for instance when regions are from a monitor with
        different pixel scaling.
        """
        left = int(self.left * scaling_factor)
        top = int(self.top * scaling_factor)
        right = int(self.right * scaling_factor)
        bottom = int(self.bottom * scaling_factor)

        return Region(left, top, right, bottom)

    def resize(self, *sizes: int) -> "Region":
        """Grow or shrink the region a given amount of pixels,
        and return the resulting copy.

        The method supports different ways to resize:

        resize(a):          a = all edges
        resize(a, b):       a = left/right, b = top/bottom
        resize(a, b, c):    a = left, b = top/bottom, c = right
        resize(a, b, c, d): a = left, b = top, c = right, d = bottom
        """
        count = len(sizes)
        if count == 1:
            left = top = right = bottom = sizes[0]
        elif count == 2:
            left = right = sizes[0]
            top = bottom = sizes[1]
        elif count == 3:
            left = sizes[0]
            top = bottom = sizes[1]
            right = sizes[2]
        elif count == 4:
            left, top, right, bottom = sizes
        else:
            raise ValueError(f"Too many resize arguments: {count}")

        left = self.left - int(left)
        top = self.top - int(top)
        right = self.right + int(right)
        bottom = self.bottom + int(bottom)

        return Region(left, top, right, bottom)

    def move(self, left: int, top: int) -> "Region":
        """Move the region relative to current position,
        and return the resulting copy.
        """
        left = self.left + int(left)
        top = self.top + int(top)
        right = left + self.width
        bottom = top + self.height

        return Region(left, top, right, bottom)

    def contains(self, element: Union[Point, "Region"]) -> bool:
        """Check if a point or region is inside this region."""
        if isinstance(element, Point):
            return (self.left <= element.x <= self.right) and (
                self.top <= element.y <= self.bottom
            )
        elif isinstance(element, Region):
            return (
                element.left >= self.left
                and element.top >= self.top
                and element.right <= self.right
                and element.bottom <= self.bottom
            )
        else:
            raise NotImplementedError("contains() only supports Points and Regions")

    def clamp(self, container: "Region") -> "Region":
        """Limit the region to the maximum dimensions defined by the container,
        and return the resulting copy.
        """
        left = max(container.left, min(self.left, container.right))
        top = max(container.top, min(self.top, container.bottom))
        right = min(container.right, max(self.right, container.left))
        bottom = min(container.bottom, max(self.bottom, container.top))

        return Region(left, top, right, bottom)



import platform
from functools import wraps


def operating_system_required(*systems):
    """Decorator to restrict method for specified operating system

    :param systems: operating systems in string format
        e.g. "Linux,Darwin", default 'Windows'
    """
    systems = systems or ["Windows"]

    def _decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if platform.system() not in systems:
                raise NotImplementedError(
                    "Keyword '%s' works only with %s operating system(s)"
                    % (f.__name__, " or ".join(systems))
                )
            else:
                return f(*args, **kwargs)

        return wrapper

    return _decorator



import contextlib
import functools
from typing import Optional

from RPA.core.windows.helpers import IS_WINDOWS

if IS_WINDOWS:
    import uiautomation as auto
    from comtypes import COMError  # noqa


class ActionNotPossible(ValueError):
    """Action is not possible for the given Control."""


class ElementNotFound(ValueError):
    """No matching elements were found."""


class WindowControlError(ValueError):
    """Matching window was not found"""


class WindowsContext:
    """Shared context for all keyword libraries."""

    def __init__(self, ctx):
        self.ctx = ctx

    @property
    def logger(self):
        return self.ctx.logger

    @property
    def current_timeout(self) -> float:
        # This value can change based on `auto.SetGlobalSearchTimeout(...)` calls.
        return auto.uiautomation.TIME_OUT_SECOND

    def _window_or_none(
        self, window: "WindowsElement"  # noqa: F821
    ) -> Optional["WindowsElement"]:  # noqa: F821
        if window and window.item:
            try:
                window.item.BoundingRectangle
            except COMError:  # pylint: disable=broad-except
                # Failure to get the bounding rectangle proves that the window doesn't
                #  exist anymore.
                return None

            return window

        return None

    @property
    def anchor(self) -> Optional["WindowsElement"]:  # noqa: F821
        return self._window_or_none(self.ctx.anchor_element)

    @property
    def window(self) -> Optional["WindowsElement"]:  # noqa: F821
        return self._window_or_none(self.ctx.window_element)

    @contextlib.contextmanager
    def set_timeout(self, timeout: Optional[float] = None):
        """Context manager that sets a custom temporary `timeout`."""
        try:
            if timeout is not None:
                auto.SetGlobalSearchTimeout(timeout)
                self.logger.info("Locator timeout set to: %f", self.current_timeout)
            yield
        finally:
            if timeout is not None:
                auto.SetGlobalSearchTimeout(self.ctx.global_timeout)
                self.logger.debug(
                    "Locator timeout set back to global value: %f", self.current_timeout
                )


def with_timeout(func):
    """Applies a temporary `timeout` over the entire decorated context method."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        timeout: Optional[float] = kwargs.pop("timeout", None)
        with self.set_timeout(timeout):
            # Do not send on purpose the `timeout` back to the original decorated
            #  function. (as the temporary timeout setting is currently active)
            return func(self, *args, **kwargs)

    return wrapper



import base64
from io import BytesIO
from pathlib import Path
from typing import Dict, List, Optional

from PIL import Image

from RPA.core.vendor.deco import keyword as method
from RPA.core.windows.context import WindowsContext
from RPA.core.windows.helpers import IS_WINDOWS, get_process_list

if IS_WINDOWS:
    import uiautomation as auto
    import win32api
    import win32con
    import win32gui
    import win32process
    import win32ui


class WindowMethods(WindowsContext):
    """Keywords for handling the Windows GUI windows"""

    @staticmethod
    def get_icon(
        filepath: Optional[str], icon_save_directory: Optional[str] = None
    ) -> Optional[str]:
        if not filepath:
            return None  # no file path to save the icon from

        # TODO: Get icons of different sizes.
        small, large = win32gui.ExtractIconEx(filepath, 0, 10)
        if len(large) <= 0:
            return None  # no icon to extract

        if len(small) > 0:
            win32gui.DestroyIcon(small[0])  # get rid of the small one

        hdc = win32ui.CreateDCFromHandle(win32gui.GetDC(0))
        ico_x = win32api.GetSystemMetrics(win32con.SM_CXICON)
        ico_y = win32api.GetSystemMetrics(win32con.SM_CYICON)
        hbmp = win32ui.CreateBitmap()
        hbmp.CreateCompatibleBitmap(hdc, ico_x, ico_y)
        hdc = hdc.CreateCompatibleDC()
        hdc.SelectObject(hbmp)
        hdc.DrawIcon((0, 0), large[0])

        executable_path = Path(filepath)
        icon_path = icon_name = f"icon_{executable_path.name}.bmp"
        if icon_save_directory:
            icon_path = str((Path(icon_save_directory) / icon_name).resolve())
        hbmp.SaveBitmapFile(hdc, icon_path)
        with Image.open(icon_path) as img:
            buffered = BytesIO()
            img.save(buffered, format="PNG")
            image_string = base64.b64encode(buffered.getvalue()).decode()
        if not icon_save_directory:
            Path(icon_path).unlink()
        return image_string

    @staticmethod
    def get_fullpath(pid: int) -> str:
        handle = win32api.OpenProcess(
            win32con.PROCESS_QUERY_LIMITED_INFORMATION, False, pid
        )
        return win32process.GetModuleFileNameEx(handle, 0)

    @method
    def list_windows(
        self,
        icons: bool = False,
        icon_save_directory: Optional[str] = None,
    ) -> List[Dict]:
        windows = auto.GetRootControl().GetChildren()
        process_list = get_process_list() if self.ctx.list_processes else {}
        win_list = []
        for win in windows:
            pid = win.ProcessId
            fullpath = None
            try:
                fullpath = self.get_fullpath(pid)
            except Exception as err:  # pylint: disable=broad-except
                self.logger.info("Open process error in `List Windows`: %s", err)
            handle = win.NativeWindowHandle
            icon_string = (
                self.get_icon(fullpath, icon_save_directory) if icons else None
            )
            rect = win.BoundingRectangle
            rectangle = (
                [rect.left, rect.top, rect.right, rect.bottom] if rect else [None] * 4
            )
            name = process_list[pid] if pid in process_list else None
            if not name and fullpath:
                name = Path(fullpath).name
            info = {
                "title": win.Name,
                "pid": pid,
                "name": name,
                "path": fullpath,
                "handle": handle,
                "icon": icon_string,
                "automation_id": win.AutomationId,
                "control_type": win.ControlTypeName,
                "class_name": win.ClassName,
                "rectangle": rectangle,
                "keyboard_focus": win.HasKeyboardFocus,
                "is_active": handle == auto.GetForegroundWindow(),
                "object": win,
            }
            if icons and not icon_string:
                self.logger.info("Icon for %s returned empty", win.Name)
            win_list.append(info)
        return win_list



from dataclasses import dataclass, asdict, fields, MISSING
from typing import Optional


@dataclass
class Locator:
    """Baseclass for a locator entry."""

    def __str__(self):
        values = []
        for field in fields(self):
            if field.default is not MISSING:
                break
            values.append(getattr(self, field.name))

        typename = NAMES[type(self)]
        return "{}:{}".format(typename, ",".join(str(value) for value in values))

    @staticmethod
    def from_dict(data):
        """Construct correct locator subclass from dictionary,
        which should contain a 'type' field and at least all
        required fields of that locator.
        """
        type_ = data.pop("type", None)
        if not type_:
            raise ValueError("Missing locator type field")

        class_ = TYPES.get(type_)
        if not class_:
            raise ValueError(f"Unknown locator type: {type_}")

        # Check for missing parameters
        required = set(
            field.name for field in fields(class_) if field.default is MISSING
        )
        missing = set(required) - set(data)
        if missing:
            raise ValueError("Missing locator field(s): {}".format(", ".join(missing)))

        # Ignore extra data
        required_or_optional = [field.name for field in fields(class_)]
        kwargs = {k: v for k, v in data.items() if k in required_or_optional}

        return class_(**kwargs)

    def to_dict(self):
        """Convert locator instance to a dictionary with type information."""
        data = {"type": NAMES[type(self)]}
        data.update(asdict(self))
        return data


@dataclass
class PointLocator(Locator):
    """Locator for absolute coordinates."""

    x: int
    y: int

    def __post_init__(self):
        self.x = int(self.x)
        self.y = int(self.y)


@dataclass
class OffsetLocator(Locator):
    """Locator for offset coordinates."""

    x: int
    y: int

    def __post_init__(self):
        self.x = int(self.x)
        self.y = int(self.y)


@dataclass
class RegionLocator(Locator):
    """Locator for area defined by coordinates."""

    left: int
    top: int
    right: int
    bottom: int

    def __post_init__(self):
        self.left = int(self.left)
        self.top = int(self.top)
        self.right = int(self.right)
        self.bottom = int(self.bottom)


@dataclass
class SizeLocator(Locator):
    """Locator for area defined by width/height."""

    width: int
    height: int

    def __post_init__(self):
        self.width = int(self.width)
        self.height = int(self.height)


@dataclass
class ImageLocator(Locator):
    """Image-based locator for template matching."""

    path: str
    confidence: Optional[float] = None
    source: Optional[str] = None  # TODO: Remove when crop is implemented

    def __post_init__(self):
        if self.confidence is not None:
            self.confidence = float(self.confidence)


@dataclass
class OcrLocator(Locator):
    """Locator for OCR-based text."""

    text: str
    confidence: Optional[float] = None
    """3-character ISO 639-2 language code. Passed to pytesseract lang parameter."""
    language: Optional[str] = None

    def __post_init__(self):
        self.text = str(self.text)
        if self.confidence is not None:
            self.confidence = float(self.confidence)


@dataclass
class BrowserLocator(Locator):
    """Browser-based locator for DOM elements."""

    strategy: str
    value: str
    source: Optional[str] = None
    screenshot: Optional[str] = None


@dataclass
class WindowsLocator(Locator):
    """Windows-based locator for windows UI elements"""

    window: str
    value: str
    version: float
    screenshot: Optional[str] = None


# Aliases for backwards compatibility, just in case.
Offset = OffsetLocator
BrowserDOM = BrowserLocator
ImageTemplate = ImageLocator
Coordinates = PointLocator

# Mapping of supported locator typenames to classes.
# Used for parsing locator literals.
TYPES = {
    "point": PointLocator,
    "offset": OffsetLocator,
    "region": RegionLocator,
    "size": SizeLocator,
    "image": ImageLocator,
    "ocr": OcrLocator,
    "browser": BrowserLocator,
    "coordinates": PointLocator,  # Backwards compatibility
    "windows": WindowsLocator,
}

# Above mapping but in reverse direction.
NAMES = {
    PointLocator: "point",
    OffsetLocator: "offset",
    RegionLocator: "region",
    SizeLocator: "size",
    ImageLocator: "image",
    OcrLocator: "ocr",
    BrowserLocator: "browser",
    PointLocator: "point",  # Backwards compatibility
    WindowsLocator: "windows",
}



from functools import wraps
from importlib import util
import inspect
import os
from typing import Any

IPYTHON_AVAILABLE = False

ipython_module = util.find_spec("IPython")
if ipython_module:
    # pylint: disable=C0415
    from IPython.display import (  # noqa
        Audio,
        display,
        FileLink,
        FileLinks,
        Image,
        JSON,
        Markdown,
        Video,
    )

    IPYTHON_AVAILABLE = True


def _get_caller_prefix(calframe):
    keyword_name = (
        calframe[1][3] if calframe[1][3] not in ["<module>", "<lambda>"] else None
    )
    if keyword_name:
        keyword_name = keyword_name.replace("_", " ").title()
        return f"Output from **{keyword_name}**"
    return ""


def print_precheck(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if not IPYTHON_AVAILABLE:
            return None
        output_level = os.getenv("RPA_NOTEBOOK_OUTPUT_LEVEL", "1")
        if output_level == "0":
            return None
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        prefix = _get_caller_prefix(calframe)
        if prefix != "":
            display(Markdown(prefix))
        return f(*args, **kwargs)

    return wrapper


@print_precheck
def notebook_print(arg=None, **kwargs) -> Any:
    """Display IPython Markdown object in the notebook

    Valid parameters are `text`, `image`, `link` or `table`.

    :param text: string to output (can contain markdown)
    :param image: path to the image file
    :param link: path to the link
    :param table: `RPA.Table` object to print
    """
    if arg and "text" in kwargs.keys():
        kwargs["text"] = f"{arg} {kwargs['text']}"
    else:
        kwargs["text"] = str(arg)
    output = _get_markdown(**kwargs)

    if output:
        display(Markdown(output))


@print_precheck
def notebook_file(filepath):
    """Display IPython FileLink object in the notebook

    :param filepath: location of the file
    """
    if filepath:
        display(FileLink(filepath))


@print_precheck
def notebook_dir(directory, recursive=False):
    """Display IPython FileLinks object in the notebook

    :param directory: location of the directory
    :param recursive: if all subdirectories should be shown also, defaults to False
    """
    if directory:
        display(FileLinks(directory, recursive=recursive))


@print_precheck
def notebook_table(table: Any, count: int = 20, columns=None):
    """Display RPA.Table or RPA.Table shaped data as IPython Markdown object in the notebook

    :param table: `RPA.Table` object to print
    :param count: How many rows of table to print
    :param columns: Names / headers of the table columns
    :param index: List of indices for table rows
    """
    # pylint: disable=C0415,E0611
    from RPA.Tables import Table, Tables  # noqa

    table = Table(table, columns)
    if count:
        table = Tables().table_head(table, count=count)
    output = _get_table_output(table)
    if output:
        display(Markdown(output))


@print_precheck
def notebook_image(image):
    """Display IPython Image object in the notebook

    :param image: path to the image file
    """
    if image:
        display(Image(image))


@print_precheck
def notebook_video(video):
    """Display IPython Video object in the notebook

    :param video: path to the video file
    """
    if video:
        display(Video(video))


@print_precheck
def notebook_audio(audio):
    """Display IPython Audio object in the notebook

    :param audio: path to the audio file
    """
    if audio:
        display(Audio(filename=audio))


@print_precheck
def notebook_json(json_object):
    """Display IPython JSON object in the notebook

    :param json_object: item to show
    """
    if json_object:
        display(JSON(json_object))


def _get_table_output(table):
    # pylint: disable=C0415,E0611
    from RPA.Tables import Tables, Table  # noqa

    output = ""
    try:
        if isinstance(table, Table):
            output = "<table class='rpafw-notebook-table'>"
            header = Tables().table_head(table, count=1)
            for row in header:
                output += "<tr>"
                for h, _ in row.items():
                    output += f"<th>{h}</th>"
                output += "</tr>"
            for row in table:
                output += "<tr>"
                for _, cell in row.items():
                    output += f"<td>{cell}</td>"
                output += "</tr>"
            output += "</table><br>"
    except ImportError:
        pass
    return None if output == "" else output


def _get_markdown(**kwargs):
    output = ""
    for key, val in kwargs.items():
        if key == "text":
            output += f"<span class='rpafw-notebook-text'>{val}</span><br>"
        if key == "image":
            output += f"<img class='rpafw-notebook-image' src='{val}'><br>"
        if key == "link":
            link_text = (val[:75] + "..") if len(val) > 75 else val
            output += f"<a class='rpafw-notebook-link' href='{val}'>{link_text}</a><br>"
        if key == "table":
            output += _get_table_output(val)
    return None if output == "" else output



import logging
import sys
from importlib.metadata import PackageNotFoundError, version as get_version

from packaging import version as version_parser


LOGGER = logging.getLogger(__name__)


def init_logger(args):
    level_param = "INFO"
    for idx, param in enumerate(args):
        if param in ("-L", "--loglevel"):
            level_param_idx = idx + 1
            if level_param_idx < len(args):
                level_param = args[level_param_idx]
                break
    robot_level = level_param.split(":", 1)[0]
    levels = {
        "TRACE": logging.DEBUG,
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARN": logging.WARNING,
        "NONE": logging.CRITICAL,
    }
    level = levels.get(robot_level, logging.INFO)
    LOGGER.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    LOGGER.addHandler(handler)


def _check_pip_version(min_version: str) -> bool:
    try:
        pip_version = get_version("pip")
    except PackageNotFoundError:
        LOGGER.debug("'pip' is not installed!")
        return False

    if version_parser.parse(pip_version) >= version_parser.parse(min_version):
        LOGGER.debug(
            "Current 'pip' version %s satisfies minimum of %s.",
            pip_version,
            min_version,
        )
        return True

    LOGGER.debug(
        "Current 'pip' version %s doesn't satisfy minimum of %s.",
        pip_version,
        min_version,
    )
    return False


def use_system_certificates():
    """Exposes native system certificate stores.

    Call this before importing anything else.
    """
    # Works with Python 3.10.12, pip 23.2.1 and above.
    py_version = (3, 10, 12)
    pip_version_str = "23.2.1"
    if sys.version_info >= py_version and _check_pip_version(pip_version_str):
        try:
            import truststore  # pylint: disable=import-outside-toplevel
        except ImportError:
            LOGGER.debug("Dependency `truststore` is not installed!")
        else:
            truststore.inject_into_ssl()
            LOGGER.info(
                "Truststore injection done, using system certificate store to validate"
                " HTTPS."
            )
            return

    LOGGER.info(
        "Truststore not in use, HTTPS traffic validated against `certifi` package."
        " (requires Python %s and 'pip' %s at minimum)",
        ".".join(str(nr) for nr in py_version),
        pip_version_str,
    )


init_logger(sys.argv[1:])



import contextlib
import functools
import logging
import os
import platform
import stat
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import urljoin

from packaging import version as version_parser
from selenium import webdriver
from selenium.webdriver.common.service import Service
from selenium.webdriver.remote.webdriver import WebDriver
from webdriver_manager.chrome import ChromeDriverManager as _ChromeDriverManager
from webdriver_manager.core.download_manager import DownloadManager
from webdriver_manager.core.driver_cache import (
    DriverCacheManager as _DriverCacheManager,
)
from webdriver_manager.core.file_manager import FileManager
from webdriver_manager.core.logger import log
from webdriver_manager.core.manager import DriverManager
from webdriver_manager.core.os_manager import (
    ChromeType,
    OSType,
    OperationSystemManager as _OperationSystemManager,
    PATTERN as _PATTERN,
)
from webdriver_manager.core.utils import (
    linux_browser_apps_to_cmd,
    read_version_from_cmd,
    windows_browser_apps_to_cmd,
)
from webdriver_manager.drivers.chrome import ChromeDriver as _ChromeDriver
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import (
    EdgeChromiumDriverManager,
    IEDriverManager as _IEDriverManager,
)
from webdriver_manager.opera import OperaDriverManager

from RPA.core.robocorp import robocorp_home


class BrowserType:
    """Constants for the browser types. (expands the internal one)"""

    MSIE = "msie"
    FIREFOX = "firefox"


PATTERN = _PATTERN.copy()
PATTERN[BrowserType.MSIE] = r"\d+\.\d+\.\d+\.\d+"


class OperationSystemManager(_OperationSystemManager):
    """Custom manager for browser version retrieval which works with explicit paths."""

    @staticmethod
    def _get_browser_version(browser_type: str, paths: List[str]) -> Optional[str]:
        common_cmds = {
            OSType.LINUX: linux_browser_apps_to_cmd(*paths),
            OSType.MAC: f"{paths[0]} --version",
            OSType.WIN: windows_browser_apps_to_cmd(
                *(
                    f"(Get-Item -Path '{path}').VersionInfo.FileVersion"
                    for path in paths
                )
            ),
        }
        cmd_mapping = {
            ChromeType.GOOGLE: common_cmds,
            ChromeType.CHROMIUM: common_cmds,
            ChromeType.MSEDGE: common_cmds,
            BrowserType.FIREFOX: common_cmds,
            BrowserType.MSIE: common_cmds,
        }
        try:
            cmd_mapping = cmd_mapping[browser_type][
                OperationSystemManager.get_os_name()
            ]
            pattern = PATTERN[browser_type]
            version = read_version_from_cmd(cmd_mapping, pattern)
            return version
        # pylint: disable=broad-except
        except Exception as exc:
            LOGGER.warning(
                "Can't read %r browser version due to: %s", browser_type, exc
            )
            return None

    def get_browser_version_from_os(
        self, browser_type: Optional[str] = None
    ) -> Optional[str]:
        if browser_type != BrowserType.MSIE:
            return super().get_browser_version_from_os(browser_type)

        # Add support for IE version retrieval from OS.
        # FIXME(cmin764, 15 Sep 2023): Remove this after fixing the issue below:
        #  https://github.com/SergeyPirogov/webdriver_manager/issues/625
        program_files = os.getenv("PROGRAMFILES", r"C:\Program Files")
        paths = [
            rf"{program_files}\Internet Explorer\iexplore.exe",
            rf"{program_files} (x86)\Internet Explorer\iexplore.exe",
        ]
        return self._get_browser_version(BrowserType.MSIE, paths=paths)

    def get_browser_version(
        self, browser_type: str, path: Optional[str] = None
    ) -> Optional[str]:
        if path:
            return self._get_browser_version(browser_type, paths=[path])

        return self.get_browser_version_from_os(browser_type)


class DriverCacheManager(_DriverCacheManager):
    """Fixes caching when retrieving an existing already downloaded webdriver."""

    def __init__(self, *args, file_manager: Optional[FileManager] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._os_system_manager = OperationSystemManager()
        self._file_manager = file_manager or FileManager(self._os_system_manager)


class ChromeDriver(_ChromeDriver):
    """Custom class which correctly obtains the chromedriver download URL."""

    def __init__(self, *args, versions_url: str, **kwargs):
        super().__init__(*args, **kwargs)

        self._versions_url = versions_url
        self._resolve_version = False

    def _get_resolved_version(self, determined_browser_version: Optional[str]) -> str:
        if determined_browser_version:
            # Resolution works with the first 3 atoms of the version, so exclude the
            #  4th one in case it exists.
            determined_browser_version = ".".join(
                determined_browser_version.split(".")[:3]
            )

        latest_release_url = (
            self._latest_release_url
            if (
                self._driver_version_to_download == "latest"
                or determined_browser_version is None
            )
            else f"{self._latest_release_url}_{determined_browser_version}"
        )
        resp = self._http_client.get(url=latest_release_url)
        return resp.text.rstrip()

    def get_driver_version_to_download(self) -> str:
        if self._resolve_version:
            return self.get_latest_release_version(resolve_version=True)

        return super().get_driver_version_to_download()

    # pylint: disable=arguments-differ
    def get_latest_release_version(self, resolve_version: bool = False) -> str:
        # This is activated for any chromedriver version.
        determined_browser_version = self.get_browser_version_from_os()
        if determined_browser_version:
            parsed_version = version_parser.parse(determined_browser_version)
            non_solveable = parsed_version.major >= 115
            solve_needed = resolve_version or len(parsed_version.release) < 4
            if non_solveable or not solve_needed:
                # Got a fully downloadable version that MAY be available, but we are
                #  not sure until we don't try it.
                log(
                    f"Get {determined_browser_version} {self._name} version for"
                    f" {self._browser_type}"
                )
                return determined_browser_version

        log(
            f"Get LATEST {self._name} version for {self._browser_type} based on"
            f" {determined_browser_version}"
        )
        return self._get_resolved_version(determined_browser_version)

    def _resolve_modern_url(
        self, version: Dict, *, driver_platform: str
    ) -> Optional[str]:
        downloads = version["downloads"]["chromedriver"]
        for dld in downloads:
            # There's also a chance for platform mismatch.
            if dld["platform"] == driver_platform:
                # Ensure the caching key and driver path matches the exact resolved
                #  version.
                self._driver_version_to_download = version["version"]
                return dld["url"]

        return None

    # pylint: disable=arguments-renamed
    def get_url_for_version_and_platform(
        self, browser_version: str, driver_platform: str
    ) -> str:
        # This is activated for chromedriver 115 or higher only.
        parse_version = lambda ver: version_parser.parse(ver).release  # noqa: E731
        parse_floating_version = lambda ver: parse_version(ver)[:3]  # noqa: E731
        parsed_floating_version = parse_floating_version(browser_version)
        resolve_modern_url = functools.partial(
            self._resolve_modern_url, driver_platform=driver_platform
        )

        response = self._http_client.get(self._versions_url)
        data = response.json()
        versions = data["versions"]

        candidates = []
        for version in versions:
            if version["version"] == browser_version:  # exact match
                url = resolve_modern_url(version)
                if url:
                    return url
            elif parsed_floating_version == parse_floating_version(version["version"]):
                candidates.append(version)

        # No exact version found, let's return the latest from the matching candidates.
        if candidates:
            latest = max(
                candidates, key=lambda candidate: parse_version(candidate["version"])
            )
            url = resolve_modern_url(latest)
            if url:
                return url

        raise Exception(f"No such driver version {browser_version} for {platform}")

    # pylint: disable=arguments-differ
    def get_driver_download_url(self, os_type: str, resolve: bool = False) -> str:
        if resolve:
            # This time we want a resolved version based on the previously parsed
            #  non-existing one on the server.
            self._resolve_version = True  # force resolve
            self._driver_version_to_download = None  # ensures resolving

        return super().get_driver_download_url(os_type)


class ChromeDriverManager(_ChromeDriverManager):
    """Custom Chrome webdriver manager which correctly downloads the chromedriver."""

    def __init__(
        self,
        driver_version: Optional[str] = None,
        name: str = "chromedriver",
        url: str = "https://chromedriver.storage.googleapis.com",
        latest_release_url: str = (
            "https://chromedriver.storage.googleapis.com/LATEST_RELEASE"
        ),
        chrome_type: str = ChromeType.GOOGLE,
        download_manager: Optional[DownloadManager] = None,
        cache_manager: Optional[DriverCacheManager] = None,
        os_system_manager: Optional[OperationSystemManager] = None,
        versions_url: str = (
            "https://googlechromelabs.github.io/chrome-for-testing/"
            "known-good-versions-with-downloads.json"
        ),
    ):
        super().__init__(
            driver_version=driver_version,
            name=name,
            url=url,
            latest_release_url=latest_release_url,
            chrome_type=chrome_type,
            download_manager=download_manager,
            cache_manager=cache_manager,
            os_system_manager=os_system_manager,
        )

        # Replace the upstream webdriver helper class with our custom auto-resolving
        #  behavior.
        self.driver = ChromeDriver(
            name=name,
            driver_version=driver_version,
            url=url,
            latest_release_url=latest_release_url,
            chrome_type=chrome_type,
            http_client=self.http_client,
            os_system_manager=os_system_manager,
            versions_url=versions_url,
        )

    def _get_driver_binary_path(self, driver: ChromeDriver) -> str:
        """Resolve the webdriver version if is not available for download."""
        try:
            return super()._get_driver_binary_path(driver)
        except ValueError:  # parsed version isn't found, let's try to resolve it
            os_type = self.get_os_type()
            resolved_url = driver.get_driver_download_url(os_type, resolve=True)
            file = self._download_manager.download_file(resolved_url)
            binary_path = self._cache_manager.save_file_to_cache(driver, file)
            return binary_path


class IEDriverManager(_IEDriverManager):
    """Custom driver manager class for IE webdriver. (running Edge in IE mode)"""

    # Forcefully download the 32bit version of the webdriver no matter the architecture
    #  of the Windows system since it is known that the 64bit version is limited and
    #  also creates issues, like freezing browser automation with Selenium sometimes.
    # https://www.selenium.dev/documentation/webdriver/browsers/internet_explorer/
    FORCE_32BIT = not os.getenv("RPA_ALLOW_64BIT_IE")

    def get_os_type(self) -> str:
        return "Win32" if self.FORCE_32BIT else super().get_os_type()


LOGGER = logging.getLogger(__name__)

DRIVER_ROOT = robocorp_home() / "webdrivers"
DRIVER_ROOT.mkdir(parents=True, exist_ok=True)
AVAILABLE_DRIVERS = {
    # Driver names taken from `webdrivermanager` and adapted to `webdriver_manager`.
    "chrome": ChromeDriverManager,
    "firefox": GeckoDriverManager,
    "gecko": GeckoDriverManager,
    "mozilla": GeckoDriverManager,
    # NOTE: Selenium 4 dropped support for Opera.
    #  (https://github.com/SeleniumHQ/selenium/issues/10835)
    "opera": OperaDriverManager,
    # NOTE: In Selenium 4 `Edge` is the same with `ChromiumEdge`.
    "edge": EdgeChromiumDriverManager,
    "chromiumedge": EdgeChromiumDriverManager,
    # NOTE: IE is discontinued and not supported/encouraged anymore.
    "ie": IEDriverManager,
}
# Control Room decides from where the webdrivers should be downloaded, but the user
#  can opt out from using our own trusted internal source and default to
#  `webdriver-manager`'s implicit locations.
_USE_EXTERNAL_WEBDRIVERS = bool(os.getenv("RPA_EXTERNAL_WEBDRIVERS"))
_SOURCE_BASE = os.getenv(
    "RC_WEBDRIVER_SOURCE", "https://downloads.robocorp.com/ext/webdrivers/"
)
_join_base = functools.partial(urljoin, _SOURCE_BASE)
_DRIVER_SOURCES = {
    "chrome": {
        "url": _join_base("chrome/v1"),
        "latest_release_url": _join_base("chrome/v1/LATEST_RELEASE"),
        "versions_url": _join_base("chrome/v2/known-good-versions-with-downloads.json"),
    },
    "firefox": {
        "url": _join_base("firefox/download"),
        "latest_release_url": _join_base("firefox/releases/latest"),
        "mozila_release_tag": _join_base("firefox/releases/tags/{0}"),
    },
    "edge": {
        "url": _join_base("edge"),
        "latest_release_url": _join_base("edge/LATEST_RELEASE"),
    },
    "ie": {
        "url": _join_base("ie/download"),
        "latest_release_url": _join_base("ie/releases"),
        "ie_release_tag": _join_base("ie/releases/tags/selenium-{0}"),
    },
}
_DRIVER_SOURCES["gecko"] = _DRIVER_SOURCES["mozilla"] = _DRIVER_SOURCES["firefox"]
_DRIVER_SOURCES["chromiumedge"] = _DRIVER_SOURCES["edge"]
# Available `WebDriver` classes in Selenium which also support automatic webdriver
#  download with `webdriver-manager`.
SUPPORTED_BROWSERS = dict(
    {name: name.capitalize() for name in AVAILABLE_DRIVERS},
    **{"chromiumedge": "ChromiumEdge"},
)
_DRIVER_PREFERENCE = {
    "Windows": ["Chrome", "Firefox", "Edge"],
    "Linux": ["Chrome", "Firefox", "Edge"],
    "Darwin": ["Chrome", "Firefox", "Edge", "Safari"],
    "default": ["Chrome", "Firefox"],
}

_OPS_MANAGER = OperationSystemManager()


def _get_browser_order_from_env() -> Optional[List[str]]:
    browsers: str = os.getenv("RPA_SELENIUM_BROWSER_ORDER", "")
    if browsers:
        return [browser.strip() for browser in browsers.split(sep=",")]

    return None  # meaning there's no env var to control the order


def get_browser_order() -> List[str]:
    """Get a list of preferred browsers based on the environment variable
    `RPA_SELENIUM_BROWSER_ORDER` if set.

    The OS dictates the order if no such env var is set.
    """
    browsers: Optional[List[str]] = _get_browser_order_from_env()
    if browsers:
        return browsers

    return _DRIVER_PREFERENCE.get(platform.system(), _DRIVER_PREFERENCE["default"])


@contextlib.contextmanager
def suppress_logging():
    """Suppress webdriver-manager logging."""
    wdm_log = "WDM_LOG"
    wdm_log_level = "WDM_LOG_LEVEL"
    old_wdm_log = os.getenv(wdm_log, "")
    old_wdm_log_level = os.getenv(wdm_log_level, "")
    try:
        os.environ[wdm_log] = str(logging.NOTSET)
        os.environ[wdm_log_level] = "0"
        yield
    finally:
        os.environ[wdm_log] = old_wdm_log
        os.environ[wdm_log_level] = old_wdm_log_level


def start(browser: str, service: Optional[Service] = None, **options) -> WebDriver:
    """Start a webdriver with the given options."""
    browser = browser.strip()
    webdriver_factory = getattr(webdriver, browser, None)
    if not webdriver_factory:
        raise ValueError(f"Unsupported Selenium browser: {browser}")

    # NOTE: It is recommended to pass a `service` rather than deprecated `options`.
    driver = webdriver_factory(service=service, **options)
    return driver


@functools.lru_cache(maxsize=1)
def _is_chromium() -> bool:
    """Detects if Chromium is used instead of Chrome no matter the platform."""
    is_browser = lambda browser_type: bool(  # noqa: E731
        _OPS_MANAGER.get_browser_version_from_os(browser_type)
    )
    with suppress_logging():
        return not is_browser(ChromeType.GOOGLE) and is_browser(ChromeType.CHROMIUM)


def _get_browser_lower(browser: str) -> str:
    browser = browser.strip()
    browser_lower = browser.lower()
    if browser_lower not in AVAILABLE_DRIVERS:
        raise ValueError(
            f"Unsupported browser {browser!r} for webdriver download!"
            f" (choose from: {', '.join(SUPPORTED_BROWSERS.values())})"
        )

    return browser_lower


def _to_manager(browser: str, *, root: Path) -> DriverManager:
    browser_lower = _get_browser_lower(browser)
    manager_factory = AVAILABLE_DRIVERS[browser_lower]
    if manager_factory == ChromeDriverManager and _is_chromium():
        manager_factory = functools.partial(
            manager_factory, chrome_type=ChromeType.CHROMIUM
        )
    if not _USE_EXTERNAL_WEBDRIVERS:
        url_kwargs = _DRIVER_SOURCES.get(browser_lower)
        if url_kwargs:
            manager_factory = functools.partial(manager_factory, **url_kwargs)
        else:
            LOGGER.warning("Can't set an internal webdriver source for %r!", browser)

    cache_manager = DriverCacheManager(root_dir=str(root))
    manager = manager_factory(
        cache_manager=cache_manager, os_system_manager=_OPS_MANAGER
    )
    driver = manager.driver
    cache = getattr(functools, "cache", functools.lru_cache(maxsize=None))
    driver.get_latest_release_version = cache(driver.get_latest_release_version)
    return manager


def _set_executable(path: str) -> None:
    st = os.stat(path)
    os.chmod(
        path,
        st.st_mode | stat.S_IXOTH | stat.S_IXGRP | stat.S_IEXEC,
    )


def download(browser: str, root: Path = DRIVER_ROOT) -> str:
    """Download a webdriver binary for the given browser and return the path to it."""
    with suppress_logging():
        manager = _to_manager(browser, root=root)
        path: str = manager.install()
    if platform.system() != "Windows":
        _set_executable(path)
    LOGGER.info("Downloaded webdriver to: %s", path)
    return path


def get_browser_version(browser: str, path: Optional[str] = None) -> Optional[str]:
    """Returns the detected browser version from OS in the absence of a given `path`."""
    browser_lower = _get_browser_lower(browser)
    chrome_type = ChromeType.CHROMIUM if _is_chromium() else ChromeType.GOOGLE
    browser_types = {
        # NOTE(cmin764, 12 Sep 2023): There's no upstream support on getting the
        #  automatically detected browser version from the OS for IE, Safari and Opera.
        #  But we introduce one here for IE only.
        "chrome": chrome_type,
        "firefox": BrowserType.FIREFOX,
        "gecko": BrowserType.FIREFOX,
        "mozilla": BrowserType.FIREFOX,
        "edge": ChromeType.MSEDGE,
        "chromiumedge": ChromeType.MSEDGE,
        "ie": BrowserType.MSIE,
    }
    browser_type = browser_types.get(browser_lower)
    if not browser_type:
        LOGGER.warning("Can't determine browser version for %r!", browser)
        return None

    return _OPS_MANAGER.get_browser_version(browser_type, path=path)



import logging
from typing import Optional

from ..vendor.robotlibcore import DynamicCore
from .action import ActionMethods
from .elements import ElementMethods
from .helpers import IS_WINDOWS
from .locators import Locator, LocatorMethods
from .window import WindowMethods

if IS_WINDOWS:
    import uiautomation as auto


class WindowsElementsMixin:
    """Windows elements mixin compatible with inspector and windows packages."""

    def __init__(self, locators_path: Optional[str] = None):
        self.logger = logging.getLogger(__name__)
        self.list_processes = True
        self.global_timeout: float = float(
            auto.uiautomation.TIME_OUT_SECOND if IS_WINDOWS else 10
        )
        self.window_element: Optional[Locator] = None
        self.anchor_element: Optional[Locator] = None

        # Register all "keyword" methods into this "library" class.
        libraries = self._get_libraries(locators_path)
        super().__init__(libraries)

    def _get_libraries(self, locators_path: Optional[str]):
        return [
            ActionMethods(self),
            ElementMethods(self),
            LocatorMethods(self, locators_path=locators_path),
            WindowMethods(self),
        ]


class WindowsElements(WindowsElementsMixin, DynamicCore):
    """Windows elements library compatible with the inspector and windows packages."""
